import asyncio
import logging
import signal

import coloredlogs
from bleak import BleakClient, BleakScanner, exc

# Set up logging with colored output
coloredlogs.install(level="DEBUG")

# Set bleak logger to INFO
logging.getLogger("bleak").setLevel(logging.INFO)



class SmartSolar(BLEDevice):
    FILTERS = [{"name": "SmartSolar "}]

    def handle_char_306b0002(self, data):
        # Get the first two bytes and interpret them as a 16-bit unsigned integer
        battery_voltage = int.from_bytes(data[:2], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        pv_voltage = int.from_bytes(data[2:4], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit signed integer
        pv_current = int.from_bytes(data[4:6], byteorder="little", signed=True)

        # Get the final byte and interpret it as an 8-bit unsigned integer
        charge_current = data[6]

        print(f"Battery voltage: {battery_voltage}")
        print(f"PV voltage: {pv_voltage}")
        print(f"PV current: {pv_current}")
        print(f"Charge current: {charge_current}")

    def handle_char_97580002(self, data):
        # Get the first two bytes and interpret them as a 16-bit signed integer
        temperature = int.from_bytes(data[:2], byteorder="little", signed=True)

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        charge_voltage = int.from_bytes(data[2:4], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        charge_current = int.from_bytes(data[4:6], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        battery_voltage = int.from_bytes(data[6:8], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit signed integer
        battery_current = int.from_bytes(data[8:10], byteorder="little", signed=True)

        # Get the next four bytes and interpret them as a 32-bit unsigned integer
        battery_capacity = int.from_bytes(data[10:14], byteorder="little")

        # Get the next four bytes and interpret them as a 32-bit signed integer
        battery_power = int.from_bytes(data[14:18], byteorder="little", signed=True)

        # Get the final two bytes and interpret them as a 16-bit unsigned integer
        yield_today = int.from_bytes(data[18:], byteorder="little")

        print(f"Temperature: {temperature}")
        print(f"Charge voltage: {charge_voltage}")
        print(f"Charge current: {charge_current}")
        print(f"Battery voltage: {battery_voltage}")
        print(f"Battery current: {battery_current}")
        print(f"Battery capacity: {battery_capacity}")
        print(f"Battery power: {battery_power}")
        print(f"Yield today: {yield_today}")

    def handle_char_97580006(self, data):
        self.log.error("Not implemented yet.")

    HANDLERS = {
        "306b0002-b081-4037-83dc-e59fcc3cdfd0": handle_char_306b0002,
        "97580002-ddf1-48be-b73e-182664615d8e": handle_char_97580002,
        "97580006-ddf1-48be-b73e-182664615d8e": handle_char_97580006,
    }






import asyncio
import logging
import signal

import coloredlogs
from bleak import BleakClient, BleakScanner, exc

# Set up logging with colored output
coloredlogs.install(level="DEBUG")

# Set bleak logger to INFO
logging.getLogger("bleak").setLevel(logging.INFO)


class BLEDevice:
    HANDLERS = []
    FILTERS = []
    PAIRING_REQUIRED = True

    def __init__(self, device, loop):
        self.device = device
        self.loop = loop
        self.log = logging.getLogger(f"Device {device.name}")
        self.log.setLevel(logging.DEBUG)

    @classmethod
    def filter(cls, device):
        for f in cls.FILTERS:
            if "name" in f:
                if device.name.startswith(f["name"]):
                    return True
            if "address" in f:
                if device.address.startswith(f["address"]):
                    return True
        return False

    async def run(self):
        self.log.info("Starting...")
        # connect to device
        async with BleakClient(self.device.address, loop=self.loop) as client:
            # read device name
            # name = await client.read_gatt_char("00002a00-0000-1000-8000-00805f9b34fb")
            # self.log.debug(f"Device name: {name.decode('utf-8')}")

            if self.PAIRING_REQUIRED:
                res = await client.pair(protection_level=3)
                self.log.debug(f"Pairing result: {res!r}")

            readable_characteristics = []

            disconnected_event = asyncio.Event()

            def disconnect_callback(client):
                self.log.info(f"Disconnected!")
                self.loop.call_soon_threadsafe(disconnected_event.set)

            client.set_disconnected_callback(disconnect_callback)

            # read characteristics
            for service in client.services:
                self.log.debug(f"Service: {service.uuid}")
                for characteristic in service.characteristics:
                    self.log.debug(f"  Characteristic: {characteristic.uuid}")

                    # check if characteristic is readable
                    self.log.debug(characteristic.properties)
                    if "read" in characteristic.properties:
                        # value = await client.read_gatt_char(characteristic.uuid)
                        # self.log.debug(f"    Value: {value}")
                        readable_characteristics.append(characteristic.uuid)

                    if "notify" in characteristic.properties:
                        self.log.debug("    Subscribing to notifications...")
                        try:
                            await client.start_notify(
                                characteristic.uuid, notify_callback
                            )
                        except Exception as e:
                            self.log.exception("Error subscribing to notifications")

            # run this task forever
            while True:
                # read all readable characteristics
                for characteristic in readable_characteristics:
                    try:
                        value = await client.read_gatt_char(characteristic)
                    except Exception as e:
                        self.log.exception(
                            "Error reading characteristic {characteristic}"
                        )
                        return
                    # print the characteristic and its value as a hex array, grouped in bytes
                    self.log.debug(
                        f"  {characteristic}: {' '.join(f'{b:02x}' for b in value)}"
                    )
                    # call the respective handler, if it exists
                    if characteristic in self.HANDLERS:
                        self.HANDLERS[characteristic](self, value)
                await asyncio.sleep(1.0)


class SmartSolar(BLEDevice):
    FILTERS = [{"name": "SmartSolar "}]

    def handle_char_306b0002(self, data):
        # Get the first two bytes and interpret them as a 16-bit unsigned integer
        battery_voltage = int.from_bytes(data[:2], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        pv_voltage = int.from_bytes(data[2:4], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit signed integer
        pv_current = int.from_bytes(data[4:6], byteorder="little", signed=True)

        # Get the final byte and interpret it as an 8-bit unsigned integer
        charge_current = data[6]

        print(f"Battery voltage: {battery_voltage}")
        print(f"PV voltage: {pv_voltage}")
        print(f"PV current: {pv_current}")
        print(f"Charge current: {charge_current}")

    def handle_char_97580002(self, data):
        # Get the first two bytes and interpret them as a 16-bit signed integer
        temperature = int.from_bytes(data[:2], byteorder="little", signed=True)

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        charge_voltage = int.from_bytes(data[2:4], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        charge_current = int.from_bytes(data[4:6], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit unsigned integer
        battery_voltage = int.from_bytes(data[6:8], byteorder="little")

        # Get the next two bytes and interpret them as a 16-bit signed integer
        battery_current = int.from_bytes(data[8:10], byteorder="little", signed=True)

        # Get the next four bytes and interpret them as a 32-bit unsigned integer
        battery_capacity = int.from_bytes(data[10:14], byteorder="little")

        # Get the next four bytes and interpret them as a 32-bit signed integer
        battery_power = int.from_bytes(data[14:18], byteorder="little", signed=True)

        # Get the final two bytes and interpret them as a 16-bit unsigned integer
        yield_today = int.from_bytes(data[18:], byteorder="little")

        print(f"Temperature: {temperature}")
        print(f"Charge voltage: {charge_voltage}")
        print(f"Charge current: {charge_current}")
        print(f"Battery voltage: {battery_voltage}")
        print(f"Battery current: {battery_current}")
        print(f"Battery capacity: {battery_capacity}")
        print(f"Battery power: {battery_power}")
        print(f"Yield today: {yield_today}")

    def handle_char_97580006(self, data):
        self.log.error("Not implemented yet.")

    HANDLERS = {
        "306b0002-b081-4037-83dc-e59fcc3cdfd0": handle_char_306b0002,
        "97580002-ddf1-48be-b73e-182664615d8e": handle_char_97580002,
        "97580006-ddf1-48be-b73e-182664615d8e": handle_char_97580006,
    }


class BlueBattery(BLEDevice):
    FILTERS = [{"name": "BlueBattery_"}]
    PAIRING_REQUIRED = False

    def handle_live(self, data):
        self.log.error("Received LIVE")

    HANDLERS = {
        "4b616912-40bd-428b-bf06-698e5e422cd9": handle_live,
    }


async def notify_callback(sender, data):
    print(f"XXXXXXXXXXXX Notification from {sender}: {data}")


KNOWN_DEVICES = (
    #SmartSolar,
    BlueBattery,
)

# default logger
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


async def scan_and_connect(tasks, loop):
    device_list = []

    # scan for devices
    devices = await BleakScanner.discover(timeout=5.0, loop=loop)
    for device in devices:
        log.debug(device.metadata)
        for k in KNOWN_DEVICES:
            if k.filter(device):
                log.info(
                    f"Found device: {device.name} ({device.address})"
                )
                device_list.append((k, device))
                break
        else:
            log.info(f"Found other device: {device.name} ({device.address})")

    # ensure discovery has finished
    await asyncio.sleep(2.0)

    if not devices:
        return

    for cls, device in device_list:
        # check if any of the running tasks already covers this device
        if device.address in tasks:
            # if task is running, skip this device
            if not tasks[device.address].done():
                log.debug(f"Task for {device.address} is already running. Skipping.")
                continue
            # if task is done, remove it from the list
            else:
                del tasks[device.address]

        # else create a new task
        log.info(f"Connecting to {device.address}...")
        instance = cls(device, loop)
        try:
            device_task = asyncio.create_task(instance.run())
        except exc.BleakError as e:
            log.exception(f"Error connecting to {device.address}")
        tasks[device.address] = device_task


tasks = {}

async def scanTask():
    loop = asyncio.get_running_loop()
    while True:
        log.info("Scanning for devices...")
        try:
            await scan_and_connect(tasks, loop)
        except exc.BleakError as e:
            log.exception("Error scanning for devices")
        await asyncio.sleep(10.0)


async def main():
    loop = asyncio.get_running_loop()
    loop.create_task(scanTask())
    
    # run until shutdown
    while True:
        await asyncio.sleep(1.0)
    

def shutdown(_, __):
    log.info("Shutting down...")
    for task in tasks.values():
        task.cancel()
    loop = asyncio.get_running_loop()
    loop.stop()


signal.signal(signal.SIGINT, shutdown)
signal.signal(signal.SIGTERM, shutdown)

log.info("Started!")
asyncio.run(main())



"""
An MQTT output plugin to publish data to an MQTT broker.
"""

import logging

import paho.mqtt.client as mqtt


class MQTTOutput:
    @staticmethod
    def add_subparser(parser):
        mqtt_parser = parser.add_parser("mqtt", help="MQTT output")
        mqtt_parser.add_argument(
            "--host",
            default="localhost",
            help="MQTT broker hostname (default: localhost)",
        )
        mqtt_parser.add_argument(
            "--port", default=1883, help="MQTT broker port (default: 1883)"
        )
        mqtt_parser.add_argument(
            "--username", default=None, help="MQTT broker username (default: None)"
        )
        mqtt_parser.add_argument(
            "--password", default=None, help="MQTT broker password (default: None)"
        )
        mqtt_parser.add_argument(
            "--topic",
            default="service/bluebattery",
            help="MQTT topic to publish to will be extended by the device UUID. (default: service/bluebattery)",
        )
        mqtt_parser.add_argument(
            "--client-id",
            default="bluebattery",
            help="MQTT client ID (default: bluebattery)",
        )

    def __init__(self, args):
        self.log = logging.getLogger("output.mqtt")
        self.topic = args.topic
        self.client = mqtt.Client(args.client_id)
        if args.username:
            self.client.username_pw_set(args.username, args.password)
        self.client.connect(args.host, args.port)
        self.client.loop_start()

        # publish to the "online" topic as last will
        self.client.will_set(f"{self.topic}/online", "0", retain=True)
        self.client.publish(f"{self.topic}/online", "1", retain=True)

    def callback(self, device, data):
        frame, output_id, output_data = data
        topic = self.topic.format(device=device)
        for key, value in output_data.items():
            topic = f"{self.topic}/{device.address}/{output_id}/{key}"
            if type(value) not in (str, int, float):
                value = str(value)
            self.log.debug(f"Publishing {topic} = {value}")
            self.client.publish(topic, value, retain=False)



from hummable.bledevice import BLEDevice
from .bbcharacteristics import BCLog, BCLive, BCSec


class BlueBattery(BLEDevice):
    FILTERS = [{"name": "BlueBattery_"}]
    PAIRING_REQUIRED = False

    CHARACTERISTICS = [BCSec, BCLive, BCLog]



import logging
import struct
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Tuple, Union

BYTE_ORDER = ">"  # > big-endian, < little-endian


log = logging.getLogger("Commands")


@dataclass
class BBValue:
    """
    Represents a value within a bluetooth bytestream (reading or writing).

    The byte representation is defined in python's struct format.

    Additionally implements ¾ as a special struct character for 24-bit signed
    integers; works only on the leftmost place in the string.
    """

    struct: str
    output_id: str
    conversion_fn: Optional[Callable] = None

    def value(self, raw_value: Union[bytes, int, float]) -> Union[int, float, str]:
        """Converts a raw_value resulting from unpacking a struct into a value usable for an application.

        Args:
            raw_value (Union[bytes, int, float]): Result of struct.unpack

        Returns:
            Union[int, float, str]: Parsed value
        """
        if self.struct == "¾":
            assert isinstance(raw_value, bytes)
            raw_value_unsigned = struct.unpack(">I", b"\x00" + raw_value)[0]
            raw_value = (
                raw_value_unsigned
                if not (raw_value_unsigned & 0x800000)
                else raw_value_unsigned - 0x1000000
            )
        if not self.conversion_fn:
            assert not isinstance(raw_value, bytes)
            return raw_value
        else:
            return self.conversion_fn(raw_value)

    def get_struct(self):
        if not self.struct == "¾":
            return self.struct
        return "3s"


class BBValueIgnore(BBValue):
    def __init__(self, byte_count: int = 1):
        self.struct = f"{byte_count}x"
        self.output_id = ""

    def value(self, raw_value):
        return None


@dataclass
class BBFrame:
    output_id: str  # note: may contain placeholders for output field values
    fields: List[BBValue]
    postprocess: Optional[Callable] = None

    def format(self):
        return BYTE_ORDER + "".join(field.get_struct() for field in self.fields)

    def process(self, characteristic, value):
        non_ignore_fields = filter(
            lambda field: type(field) is not BBValueIgnore, self.fields
        )

        # field values, raw from the struct unpacking
        raw_values = zip(non_ignore_fields, struct.unpack_from(self.format(), value))

        """
        Frames with multiple sub-frames may contain the same field name twice. Once
        the same field name has been observed twice, we emit a sub-frame result and
        continue with the rest of the frame.
        """
        output = {}

        for field, raw_value in raw_values:
            # existing field name indicates: begin of new sub-frame! emit old frame first
            if field.output_id in output:
                if self.postprocess:
                    self.postprocess(characteristic, output)
                yield (self, self.output_id.format(**output), output)
            # existing field value will be overwritten as necessary
            output[field.output_id] = field.value(raw_value)

        if self.postprocess:
            self.postprocess(characteristic, output)
        yield (self, self.output_id.format(**output), output)


@dataclass
class BBFrameTypeSwitch:
    index_byte: str
    frame_types: Dict[Tuple[int, ...], BBFrame]

    def process(self, characteristic, value):
        index_value = struct.unpack_from(BYTE_ORDER + self.index_byte, value)
        if index_value not in self.frame_types:
            raise Exception(f"Frame type not found: {index_value!r}")
        log.debug(f"Selected index: {index_value}.")

        yield from self.frame_types[index_value].process(characteristic, value)



import asyncio 
from concurrent.futures import CancelledError

from . import frametypes
from .commands import BBFrameTypeSwitch
from hummable.characteristics import ReadPeriodicCharacteristic


class BCLog(ReadPeriodicCharacteristic):
    """
    This characteristic is used to read the log entries from the battery computer.
    Each read access auto increments the day counter until current day is reach, then it wraps around.
    Reading starts at day first available day.
    First all type 0x00 entries are send, then all type 0x01 frames etc.
    Note: Reading "sec" characteristics resets read pointer to earliest available log entry.
    """

    UUID = "4b616907-40bd-428b-bf06-698e5e422cd9"
    PERIOD = 1
    WAIT_BETWEEN = 60 * 60  # once per hour
    MAX_LOG_FRAMES = 60 * 3  # 60 days, three types of log entries
    INITIAL_WAIT = 30

    async def read_periodically(self):
        self.log.debug(f"Starting periodic read of {self.UUID}")
        try:
            # wait some time so that sec has been read
            await asyncio.sleep(self.INITIAL_WAIT)
            while True:
                self.log.debug(f"Starting new readout of logs from {self.UUID}")
                self.reset_log_info()
                while True:
                    # read characteristic
                    data = await self.client.read_gatt_char(self.UUID)
                    self.log.debug(f"Read {self.UUID}: {' '.join(f'{b:02x}' for b in data)}")

                    parsed = self.parse(data)

                    wrapped = False
                    for frame in parsed:
                        self.log.debug(f"Parsed frame: {frame}")
                        self.output_callback(frame)
                        if self.check_log_has_wrapped(frame):
                            wrapped = True
                            break

                    if wrapped:
                        self.log.debug("Log has wrapped, stopping readout")
                        break

                    # wait a second before continuing reading
                    await asyncio.sleep(self.PERIOD)
                await asyncio.sleep(self.WAIT_BETWEEN)
        except CancelledError:
            self.log.debug("Task cancelled")
        except Exception as e:
            self.log.exception("Error reading characteristic")

    def reset_log_info(self):
        self.log_info = {
            "first_day_seen": None,
            "first_frametype_seen": None,
            "frames_seen": 0,
        }

    def check_log_has_wrapped(self, frame):
        frametype, measurement, values = frame
        self.log_info["frames_seen"] += 1
        if self.log_info["frames_seen"] == self.MAX_LOG_FRAMES:
            return True
            
        if not self.log_info["first_frametype_seen"]:
            self.log_info["first_frametype_seen"] = frametype
            self.log_info["first_day_seen"] = values["day_counter"]

        elif (
            frametype == self.log_info["first_frametype_seen"]
            and values["day_counter"] == self.log_info["first_day_seen"]
        ):
            # finished reading log entries
            return True
        return False

    def parse(self, data):
        yield from BBFrameTypeSwitch(
        "36xB",  # 36th byte is the frame type indicator
        {
            (0x00,): frametypes.LogEntryDaysFrame,
            (0x01,): frametypes.LogEntryFrameOld,
            (0x02,): frametypes.LogEntryFrameNew,
        },
        ).process(self, data)



class BCSec(ReadPeriodicCharacteristic):
    """
    time of day in seconds, after power-up it starts with 0, needs to be set after initial connection to correct time.
    Log entry is generated when seconds reach 86400 (24 Hours) and seconds are reset to 0.
    """

    UUID = "4b616901-40bd-428b-bf06-698e5e422cd9"
    PERIOD = 10 * 60  # once every 10 minutes

    def parse(self, data):
        yield from frametypes.SecFrame.process(self, data)


class BCLive(ReadPeriodicCharacteristic):
    UUID = "4b616912-40bd-428b-bf06-698e5e422cd9"
    PERIOD = 1

    def parse(self, data):
        yield from BBFrameTypeSwitch(
            "BB",  # first two bytes indicate frame type
            {
                # byte 0: type
                # byte 1: length
                (0x00, 0x07): frametypes.BCLiveMeasurementsFrame,
                (0x00, 0x09): frametypes.BCLiveMeasurementsFrameExtended,
                (0x01, 0x09): frametypes.BCSolarChargerEBLFrame,
                (0x01, 0x0B): frametypes.BCSolarChargerStandardFrame,
                (0x01, 0x0C): frametypes.BCSolarChargerExtendedFrame,
                (0x02, 0x10): frametypes.BCBatteryComputer1Frame,
                (0x03, 0x0F): frametypes.BCBatteryComputer2Frame,
                (0x04, 0x01): frametypes.BCIntradayLogEntryFrame,
                (0x05, 0x0A): frametypes.BCBoosterDataFrame,
                (0x05, 0x04): frametypes.BCNoBoosterDataFrame,
            },
        ).process(self, data)



def cnv_mV_to_V(mV):
    return mV / 1000


cnv_mW_to_W = cnv_mV_to_V


def cnv_10mV_to_V(_10mV):
    return _10mV / 100


cnv_10mA_to_A = cnv_10mV_to_V

def cnv_10mAh_to_Ah(_10mAh):
    return _10mAh / 100 

def cnv_mA_to_A(mA):
    return mA / 1000


def cnv_100mA_to_A(_100mA):
    return _100mA / 10


cnv_100mV_to_V = cnv_100mA_to_A


def cnv_neg_100mA_to_A(_100mA):
    return -_100mA / 10


def cnv_bb_temp_to_deg_c(bb_temp):
    return (bb_temp - 0x8000) / 100


def cnv_solar_status(status):
    return {0: "active", 1: "standby", 2: "reduced"}.get(status, "unknown")



import logging
import signal
import argparse
import asyncio

import coloredlogs

from .bbdevice import BlueBattery
from .output.log import LogOutput
from .output.mqtt import MQTTOutput
from hummable.scanner import Scanner


KNOWN_DEVICES = (BlueBattery,)

def run():
    # Parse command line arguments, use subparsers from log and mqtt.

    parser = argparse.ArgumentParser(description="BlueBattery Service")
    subparsers = parser.add_subparsers(dest="output", help="Output")
    LogOutput.add_subparser(subparsers)
    MQTTOutput.add_subparser(subparsers)

    # let user define the log level, default is INFO

    parser.add_argument(
        "--log-level",
        default="INFO",
        help="Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
    )

    args = parser.parse_args()

    # set log level
    # Set up logging with colored output
    coloredlogs.install(level=args.log_level)

    # Set bleak logger to INFO
    logging.getLogger("bleak").setLevel(logging.INFO)

    # Create output depending on which subparser was used.
    if args.output == "log":
        output = LogOutput(args)
    elif args.output == "mqtt":
        output = MQTTOutput(args)
    else:
        raise ValueError("Please specify an output method.")

    scanner = Scanner(output.callback, KNOWN_DEVICES)

    # default logger
    log = logging.getLogger(__name__)
    log.setLevel(logging.DEBUG)

    signal.signal(signal.SIGINT, scanner.shutdown)
    signal.signal(signal.SIGTERM, scanner.shutdown)

    log.info("Started!")
    asyncio.run(scanner.run())


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bluebattery', 'bluebattery.output']

package_data = \
{'': ['*']}

install_requires = \
['bleak>=0.19.5,<0.20.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'hummable>=0.1.0,<0.2.0',
 'paho-mqtt>=1.6.1,<2.0.0',
 'py-flags>=1.1.4,<2.0.0']

entry_points = \
{'console_scripts': ['bb_cli = bluebattery.cli:run']}

setup_kwargs = {
    'name': 'bluebattery-py',
    'version': '1.0.1',
    'description': '',
    'long_description': '# bluebattery.py\n\nSoftware for interacting with the [BlueBattery](https://www.blue-battery.com/) line of battery computers for RVs.\n\nFeatures:\n\n- [x] Reading of periodically sent measurements\n- [x] Publishes value to an MQTT broker\n- [x] Auto-discovery of BB devices\n- [x] Access to stored logs\n- [ ] Modification of device settings\n- [ ] Firmware update\n\n## Changelog\n\n2023-01-14: Moved to bleak library. This should make the software much more reliable. Multiple devices are now supported and the MQTT topic has changed to include the device address.\n\n## Installation\n\n```\nsudo apt-get install python3-pip\n# log out and in again to apply new environment variables \npip3 install git+https://github.com/danielfett/bluebattery.py.git\n```\n\n## Setting up a Systemd Service\n\nSee the [systemd service](assets/bb.service) file for details.\n\n\n## Reading values from the command line\n\n```\n$ bb_cli log\n```\n\nIf you want to see more details of what is going on, use the debug flag:\n\n```\n$ bb_cli --log-level DEBUG log\n```\n\n## Publishing values to an MQTT server\n\nIf you want to use the MQTT features, start the MQTT publisher using\n\n```\n$ bb_cli mqtt\n```\n\nAppend `--help` to see the configuration options.\n\nThis is an example of the values published to the MQTT broker:\n\n```\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/max_solar_current_day_A 0.0\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/max_solar_watt_day_W 0.0\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/solar_charge_day_Ah 0.0\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/solar_energy_day_Wh 0\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/solar_charger_status 1\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/solar_module_voltage_V 0.0\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/solar_charger_ext/relay_status RelayStatus()\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/battery_charge_Ah 158.48\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/state_of_charge_percent 83.4\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/max_battery_current_day_A 0.0\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/min_battery_current_day_A -1.1\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/max_battery_charge_day_Ah 16.16\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/min_battery_charge_day_Ah 15.84\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/max_battery_voltage_day_V 12.54\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/battery_comp_1/min_battery_voltage_day_V 12.5\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/info/battery_voltage_V 12.61\nservice/bluebattery/FC:45:C3:CA:FF:EE/live/info/starter_battery_voltage_V 12.43\n```\n\n\n## Troubleshooting\n\nDepending on your environment, you may need to enable BLE first or to set up your linux user to allow using BLE:\n\n### Enabling Bluetooth LE\n\nIf the above command does not work out-of-the-box, you might have to enable Bluetooth Low-Energy. \n\nOn Ubuntu, add the following two lines at the bottom of `/etc/bluetooth/main.conf`:\n\n```\nEnableLE=true\nAttributeServer=true\n```\n\nThen restart bluetooth: `sudo service bluetooth restart`\n\n### On the Raspberry Pi\n\nThis software works on a Raspberry Pi and was tested with the built-in bluetooth device. To use the software as the user `pi` (recommended!), you need to make the dbus policy changes [described here](https://www.raspberrypi.org/forums/viewtopic.php?t=108581#p746917).\n',
    'author': 'Daniel Fett',
    'author_email': 'fett@danielfett.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging

class LogOutput:
    @staticmethod
    def add_subparser(parser):
        parser.add_parser("log", help="Log output")

    def __init__(self, args):
        self.log = logging.getLogger("output.log")

    def callback(self, device, data):
        frame, output_id, output_data = data
        self.log.info(f"{device.address} | {output_id}: {output_data}")



from flags import Flags

from . import conversions as cnv
from .commands import (
    BBFrame,
    BBValue,
    BBValueIgnore,
)

SecFrame = BBFrame(output_id="sec", fields=[BBValue("i", "time_of_day_s")])


def add_reverse_day_counter(characteristic, output_values):
    """
    Frame postprocessing function to create a negative day counter (today=0, yesterday=1, etc.).
    """

    if "max_day_count" in output_values:
        characteristic.max_days_observed = output_values["max_day_count"]

    output_values["days_ago"] = (
        output_values["max_day_count"] - output_values["day_counter"]
    )


LogEntryDaysFrame = BBFrame(
    output_id="log/day/-{days_ago}",
    fields=[
        # bytes 0-1: big endian 16 bit mWh per day**
        BBValue("H", "Wh_day", cnv.cnv_mW_to_W),
        # bytes 2-3: big endian 16-bit max mW per Day**
        BBValue("H", "max_W_day", cnv.cnv_mW_to_W),
        # bytes 4-5: big endian 16-bit min mV per Day
        BBValue("H", "min_V_day", cnv.cnv_mV_to_V),
        # bytes 6-7: big endian 16-bit max mV per Day
        BBValue("H", "max_V_day", cnv.cnv_mV_to_V),
        # bytes 8-9: big endian 16-bit charge time in minutes per Day**
        BBValue("H", "charge_minutes_day"),
        # bytes 10-11: big endian 16-bit day counter for this log entry, counting up from 0
        BBValue("H", "day_counter"),
        # bytes 12-13: big endian 16-bit current day (max day count)
        BBValue("H", "max_day_count"),
        # bytes 14-15: big endian 16-bit Solar charge in 100 mAh**
        BBValue("H", "solar_charge_Ah", cnv.cnv_100mA_to_A),
        # bytes 16-17: big endian 16-bit SOC in 100 mAh
        BBValue("H", "state_of_charge_Ah", cnv.cnv_100mA_to_A),
        # bytes 18-19: big endian 16-bit max current 100 mA
        BBValue("H", "max_current_A", cnv.cnv_100mA_to_A),
        # bytes 20-21: big endian 16-bit min current -100 mA
        BBValue("H", "min_current_A", cnv.cnv_neg_100mA_to_A),
        # bytes 22-23: big endian 16-bit max SOC 100 mAh
        BBValue("H", "max_state_of_charge_Ah", cnv.cnv_100mA_to_A),
        # bytes 24-25: big endian 16-bit min SOC 100 mAh
        BBValue("H", "min_state_of_charge_Ah", cnv.cnv_100mA_to_A),
        # bytes 26-27: big endian 16-bit binary offset max Temperature (x - 0x8000) / 100 °C
        BBValue("H", "max_temperature_deg_C", cnv.cnv_bb_temp_to_deg_c),
        # bytes 28-29: big endian 16-bit binary offset min Temperature (x - 0x8000) / 100 °C
        BBValue("H", "min_temperature_deg_C", cnv.cnv_bb_temp_to_deg_c),
        # bytes 30-31: big endian 16-bit total external charge per day 100 mAh**
        BBValue("H", "total_external_charge_day_Ah", cnv.cnv_100mA_to_A),
        # bytes 32-33: big endian 16-bit total discharge per day 100 mAh
        BBValue("H", "total_discharge_Ah", cnv.cnv_100mA_to_A),
        # bytes 34-35: big endian 16-bit total charge per day 100 mAh
        BBValue("H", "total_charge_day_Ah", cnv.cnv_100mA_to_A),
        # byte 36: 8-bit frame type 0x00
        BBValueIgnore(),
        # bytes 37-38: big endian 16-bit total booster charge per day 100 mAh (V304)**
        BBValue("H", "total_booster_charge_day_Ah", cnv.cnv_100mA_to_A),
    ],
    postprocess=add_reverse_day_counter,
)

# available starting with Version V2xx, not anymore supported starting V306
LogEntryFrameOld = BBFrame(
    output_id="log/day/-{days_ago}/extended",
    fields=[
        # bytes 0-1: 16-bit day counter (relative to current day in frame type 0x00)
        BBValue("H", "day_counter"),
        # bytes 2-3: 16-bit wall time in seconds/2
        BBValue("H", "wall_time", lambda value: value * 2),
        # bytes 4-5: 16-bit average battery voltage mV
        BBValue("H", "avg_battery_voltage_V", cnv.cnv_mV_to_V),
        # bytes 6-7: 16-bit average solar current mA
        BBValue("H", "avg_solar_current_A", cnv.cnv_mA_to_A),
        # bytes 8: solar charger status: aktiv, standby, reduced
        BBValue("B", "solar_charger_status", cnv.cnv_solar_status),
        # bytes 9-10: 16-bit average battery current 100 mA
        BBValue("H", "avg_battery_current_A", cnv.cnv_100mA_to_A),
        # bytes 11-12: 16-bit battery SOC 100 mAh
        BBValue("H", "battery_state_of_charge_A", cnv.cnv_mA_to_A),
        # bytes 13+ 0-1: 16-bit day counter (relative to current day in frame type 0x00)
        BBValue("H", "day_counter"),
        # bytes 13+ 2-3: 16-bit wall time in seconds/2
        BBValue("H", "wall_time", lambda value: value * 2),
        # bytes 13+ 4-5: 16-bit average battery voltage mV
        BBValue("H", "avg_battery_voltage_V", cnv.cnv_mV_to_V),
        # bytes 13+ 6-7: 16-bit average solar current mA
        BBValue("H", "avg_solar_current_A", cnv.cnv_mA_to_A),
        # bytes 13+ 8: solar charger status: aktiv, standby, reduced
        BBValue("B", "solar_charger_status", cnv.cnv_solar_status),
        # bytes 13+ 9-10: 16-bit average battery current 100 mA
        BBValue("H", "avg_battery_current_A", cnv.cnv_100mA_to_A),
        # bytes 13+ 11-12: 16-bit battery SOC 100 mAh
        BBValue("H", "battery_state_of_charge_A", cnv.cnv_mA_to_A),
        # bytes 26-35: reserved
        BBValueIgnore(10),
        # byte 36: 8-bit frame type 0x01
        BBValueIgnore(),
    ],
    postprocess=add_reverse_day_counter,
)

LogEntryFrameNew = BBFrame(
    output_id="log/day/-{days_ago}/extended",
    fields=[
        # bytes 0-1: 16-bit day counter (relative to current day in frame type 0x00)
        BBValue("H", "day_counter"),
        # bytes 2-3: 16-bit wall time in seconds/2
        BBValue("H", "wall_time", lambda value: value * 2),
        # bytes 4-5: 16-bit average battery voltage mV
        BBValue("H", "avg_battery_voltage_V", cnv.cnv_mV_to_V),
        # bytes 6-7: 16-bit average solar current mA
        BBValue("H", "avg_solar_current_A", cnv.cnv_mA_to_A),
        # bytes 8: solar charger status: aktiv, standby, reduced
        BBValue("B", "solar_charger_status", cnv.cnv_solar_status),
        # bytes 9-10: 16-bit average battery current 100 mA
        BBValue("H", "avg_battery_current_A", cnv.cnv_100mA_to_A),
        # bytes 11-12: 16-bit battery SOC 100 mAh
        BBValue("H", "battery_state_of_charge_A", cnv.cnv_mA_to_A),
        # bytes 13-14: 16-bit booster average input voltage 10mV (starter)**
        BBValue("H", "avg_booster_input_voltage_V", cnv.cnv_10mV_to_V),
        # bytes 15-16: signed 16-bit booster average current 100 mA**
        BBValue("h", "avg_booster_current_A", cnv.cnv_100mA_to_A),
        # bytes 17+ 0-1: 16-bit day counter (relative to current day in frame type 0x00)
        BBValue("H", "day_counter"),
        # bytes 17+ 2-3: 16-bit wall time in seconds/2
        BBValue("H", "wall_time", lambda value: value * 2),
        # bytes 17+ 4-5: 16-bit average battery voltage mV
        BBValue("H", "avg_battery_voltage_V", cnv.cnv_mV_to_V),
        # bytes 17+ 6-7: 16-bit average solar current mA
        BBValue("H", "avg_solar_current_A", cnv.cnv_mA_to_A),
        # bytes 17+ 8: solar charger status: aktiv, standby, reduced
        BBValue("B", "solar_charger_status", cnv.cnv_solar_status),
        # bytes 17+ 9-10: 16-bit average battery current 100 mA
        BBValue("H", "avg_battery_current_A", cnv.cnv_100mA_to_A),
        # bytes 17+ 11-12: 16-bit battery SOC 100 mAh
        BBValue("H", "battery_state_of_charge_A", cnv.cnv_mA_to_A),
        # bytes 17+ 13-14: 16-bit booster average input voltage 10mV (starter)**
        BBValue("H", "avg_booster_input_voltage_V", cnv.cnv_10mV_to_V),
        # bytes 17+ 15-16: signed 16-bit booster average current 100 mA**
        BBValue("h", "avg_booster_current_A", cnv.cnv_100mA_to_A),
        # bytes 34-35: reserved
        BBValueIgnore(2),
        # byte 36: 8-bit frame type 0x01
        BBValueIgnore(),
    ],
    postprocess=add_reverse_day_counter,
)


BCLiveMeasurementsFrame = BBFrame(
    output_id="live/measurement",
    fields=[
        # byte 0: type
        # byte 1: length
        BBValueIgnore(2),
        # 2 bytes (17) value battery voltage in mV
        BBValue("H", "battery_voltage_V", cnv.cnv_mV_to_V),
        # 2 bytes (18) value solar charge current in mA**
        BBValue("H", "solar_charge_current_A", cnv.cnv_mA_to_A),
        # 3 bytes (00) value Battery Current in mA
        BBValue("¾", "battery_current_A", cnv.cnv_mA_to_A),
    ],
)

BCLiveMeasurementsFrameExtended = BBFrame(
    output_id="live/measurement_ext",
    fields=BCLiveMeasurementsFrame.fields
    + [
        # 2 bytes optional heap size in bytes
        BBValue("H", "heap_size_bytes"), 
    ],
)

BCSolarChargerEBLFrame = BBFrame(
    output_id="live/solar_charger_ebl",
    fields=[
        # byte 0: type
        # byte 1: length
        BBValueIgnore(2),
        # 2 bytes (09) value Solar max Current per day in mA
        BBValue("H", "max_solar_current_day_A", cnv.cnv_mA_to_A),
        # 2 bytes (10) value Solar max Watt per day in 1W  use 95% efficiency
        BBValue("H", "max_solar_watt_day_W", lambda value: value / 0.95),
        # 2 bytes (19) value solar charge in 10mAh (*)
        BBValue("H", "solar_charge_day_Ah", cnv.cnv_10mAh_to_Ah),
        # 2 bytes (20) value solar energy in Wh
        BBValue("H", "solar_energy_day_Wh"),
        # 1 byte (21) status solar charger (*) bit 7 indicates sleep
        BBValue("B", "solar_charger_status"),
    ],
)

BCSolarChargerStandardFrame = BBFrame(
    output_id="live/solar_charger",
    fields=BCSolarChargerEBLFrame.fields
    + [
        # 2 bytes (22) value solar PV module voltage in 10 mV (*)
        BBValue("H", "solar_module_voltage_V", cnv.cnv_10mV_to_V),
    ],
)


class RelayStatus(Flags):
    enabled = ()
    trigger_soc = ()
    trigger_board_voltage = ()
    trigger_starter_voltage = ()
    trigger_temperature = ()
    trigger_solar_current = ()
    trigger_time = ()
    reserved = ()


BCSolarChargerExtendedFrame = BBFrame(
    output_id="live/solar_charger_ext",
    fields=BCSolarChargerStandardFrame.fields
    + [
        # 1 byte  Relais Status (BB-X >= V407)
        #  bit 0: 0:off / 1:on
        #  bit 1: 1:SOC
        #  bit 2: 1:Board Voltage
        #  bit 3: 1:Starter Voltage
        #  bit 4: 1:Temperature
        #  bit 5: 1:Solar Current
        #  bit 6: 1:Time
        #  bit 7: reserved
        BBValue("B", "relay_status", RelayStatus),
    ],
)

BCBatteryComputer1Frame = BBFrame(
    output_id="live/battery_comp_1",
    fields=[
        # byte 0: type
        # byte 1: length
        BBValueIgnore(2),
        # 2 bytes (01) value Battery Charge in 10mAh (*)
        BBValue("H", "battery_charge_Ah", cnv.cnv_10mAh_to_Ah),
        # 2 bytes (02) value SOC in 0.1% steps
        BBValue("H", "state_of_charge_percent", lambda x: x / 10),
        # 2 bytes (03) value Battery max Current per day in 10mA (*)
        BBValue("H", "max_battery_current_day_A", cnv.cnv_10mA_to_A),
        # 2 bytes (04) value Battery min Current per day in 10mA (*)
        BBValue("H", "min_battery_current_day_A", lambda x: x / -10),
        # 2 bytes (05) value Battery max Charge per day in 10mAh
        BBValue("H", "max_battery_charge_day_Ah", cnv.cnv_10mAh_to_Ah),
        # 2 bytes (06) value Battery min Charge per day in 10mAh
        BBValue("H", "min_battery_charge_day_Ah", cnv.cnv_10mAh_to_Ah),
        # 2 bytes (07) value Battery max Voltage per day in mV
        BBValue("H", "max_battery_voltage_day_V", cnv.cnv_mV_to_V),
        # 2 bytes (08) value Battery min Voltage per day in mV
        BBValue("H", "min_battery_voltage_day_V", cnv.cnv_mV_to_V),
    ],
)

BCBatteryComputer2Frame = BBFrame(
    output_id="live/battery_comp_2",
    fields=[
        # byte 0: type
        # byte 1: length
        BBValueIgnore(2),
        # 2 bytes (11) value Temperature in °C binary offset
        BBValue("H", "temperature_deg_C", cnv.cnv_bb_temp_to_deg_c),
        # 2 bytes (12) value min Temperature per day in °C binary offset
        BBValue("H", "min_temperature_deg_C", cnv.cnv_bb_temp_to_deg_c),
        # 2 bytes (13) value max Temperature per day in °C binary offset
        BBValue("H", "max_temperature_deg_C", cnv.cnv_bb_temp_to_deg_c),
        # 3 bytes (14) value summed total charge per day in 32/225 mAh (*)
        # TODO: unsigned
        BBValue("¾", "total_charge_day_Ah", lambda x: (x * (32 / 225)) / 1000),
        # 3 bytes (15) value summed total discharge per day in 32/225 mAh (*)
        # TODO: unsigned
        BBValue("¾", "total_discharge_day_Ah", lambda x: (x * (32 / 225)) / 1000),
        # 3 bytes (16) value summed total external charge per in 32/225 mAh (*)
        # TODO: unsigned
        BBValue("¾", "total_external_charge_day_Ah", lambda x: (x * (32 / 225)) / 1000),
    ],
)

BCIntradayLogEntryFrame = BBFrame(
    output_id="live/intraday_log",
    fields=[
        # byte 0: type
        # byte 1: length
        BBValueIgnore(2),
        # byte 2: record #
        BBValue("B", "record_number"),
    ],
)

BCNoBoosterDataFrame = BBFrame(
    output_id="live/info",
    fields=[
        # byte 0: type
        # byte 1: length
        BBValueIgnore(2),
        # 2 bytes value output voltage in 10 mV (Board Battery)**
        BBValue("H", "battery_voltage_V", cnv.cnv_10mV_to_V),
        # 2 bytes value input voltage in 10 mV (Starter Battery)**
        BBValue("H", "starter_battery_voltage_V", cnv.cnv_10mV_to_V),
    ],
)

BCBoosterDataFrame = BBFrame(
    output_id="live/booster",
    fields=BCNoBoosterDataFrame.fields
    + [
        # 2 bytes signed value charge current in 100 mA (Booster current)
        BBValue("H", "booster_charge_current_A", cnv.cnv_100mA_to_A),
        # 1 byte status booster
        BBValue("B", "booster_status"),
        # 3 bytes value summed total booster charge per day in 256/18000 Ah
        # TODO: unsigned
        BBValue("¾", "total_booster_charge_day_Ah", lambda x: x * (256 / 18000)),
    ],
)
